import requests

from . import logging_manager
from . import custom_exceptions
from . import json_parser

# User-Agent header required by Wikimedia API policy
# https://meta.wikimedia.org/wiki/User-Agent_policy
HEADERS = {
    'User-Agent': 'wikiquotes/1.4 (https://github.com/FranDepascuali/wikiquotes-python-api)'
}

def request_quote_of_the_day_page(language):
    return _request_via_scrapping(language.quote_of_the_day_url, language)

def request_quotes_page(title, language):
    answer = _request_quotes_via_api(title, language.base_url)
    if answer is None:
        raise custom_exceptions.PageNotFoundException()

    return json_parser.quotes_from_json(answer)

def request_titles(title, language):
    answer = None
    search_result = title

    while not isinstance(search_result, list):
        answer = _search_for_correct_title_via_api(search_result, language.base_url)
        search_result = json_parser.correct_title_from_json(answer)

    return search_result

def _search_for_correct_title_via_api(title, base_url):
    # https://en.wikipedia.org/w/api.php?action=query&list=search&srsearch=shakespeare
    return _request_via_api(base_url, action = 'query', list = 'search', srsearch = title, format = 'json')

def _request_quotes_via_api(title, base_url):
    return _request_via_api(base_url, titles = title, action = 'query', prop = 'extracts', format = 'json', redirects = True)

def _request_via_api(base_url, titles = None, action = None, prop = None, format = None, list = None, srsearch = None, redirects = None):
    def set_value(dictionary, key, value):
        if value:
            dictionary[key] = value

    parameters = {}
    set_value(parameters, 'action', action)
    set_value(parameters, 'prop', prop)
    set_value(parameters, 'format', format)
    set_value(parameters, 'redirects', redirects)
    set_value(parameters, 'titles', titles)
    set_value(parameters, 'list', list)
    set_value(parameters, 'srsearch', srsearch)

    request = requests.get(base_url, params = parameters, allow_redirects = redirects, headers = HEADERS)

    if format == "json":
        answer = request.json()
    else:
        # logging_manager.error("Incorrect format (json expected)", exc_info=True)
        raise custom_exceptions.IncorrectAPIFormatException()

    return answer

# Use this if it can't be achieved by _request_via_api (because _request_via_api solves redirects automatically)
def _request_via_scrapping(page, language):
    request = requests.get(page, headers = HEADERS)
    # logging_manager.info("Requesting via scrapping: {}".format(page))
    return request.content



base_url = "https://en.wikiquote.org/w/api.php"
quote_of_the_day_url = "https://en.wikiquote.org/wiki/Wikiquote:Quote_of_the_day"

def quote_of_the_day_parser(html):
    quote_and_author = (html.table)("td")[2]("td")

    quote = quote_and_author[0].text.strip()
    author = quote_and_author[1].text.replace("~", "").strip()

    return (quote, author)

non_quote_sections = ['cast', 'see also', 'external links', 'misquoted', 'misattributed', "about"]



from functools import wraps
import os.path
import shutil

def ensure_file_exists(file_path):
    if not file_exists(file_path):
        raise IOError(f'File does not exist: {file_path}')

def ensure_directory_exists(directory_path):
    if not directory_exists(directory_path):
        raise IOError(f'Directory does not exist: {directory_path}')

def ensure_file_not_exists(file_path):
    if file_exists(file_path):
        raise IOError(f'File already exist: {file_path}')

def ensure_directory_not_exists(directory_path):
    if directory_exists(directory_path):
        raise IOError(f'Directory already exist: {directory_path}')

def file_exists(file_path):
    return os.path.exists(file_path) and os.path.isfile(file_path)

def directory_exists(directory_path):
    return os.path.exists(directory_path) and os.path.isdir(directory_path)

def create_directory_if_not_exists(directory_path):
    if not directory_exists(directory_path):
        os.makedirs(directory_path)
        return

def delete_directory(directory_path):
    if directory_exists(directory_path):
        shutil.rmtree(directory_path, ignore_errors=True)

def delete_file(file_path):
    if file_exists(file_path):
        os.remove(file_path)

def last_component_of(file_path):
    return os.path.basename(file_path)

def directory_of(file_path):
    return os.path.dirname(file_path)

def list_relative_files_with_extension(directory, extension):
    return filter(lambda filename: filename.endswith(extension), os.listdir(directory))

def list_absolute_files_with_extension(directory, extension):
    relative_paths = list_relative_files_with_extension(directory, extension)
    return map(lambda relative_path: join(directory, relative_path), relative_paths)

def join(baseurl, path_to_append):
    ensure_directory_exists(baseurl)
    return os.path.join(baseurl, path_to_append)



from .wikiquotes_api import search, get_quotes, quote_of_the_day, random_quote, supported_languages

try:
    from importlib.metadata import version
    __version__ = version("wikiquotes")
except Exception:
    __version__ = "unknown"



import random
import os

from .managers import api_manager
from .managers import html_manager
from .managers import language_manager
from .managers import custom_exceptions

def search(author, raw_language):
    # logging_manager.info("API intent: {} author: {} raw_language: {}".format("search", author, raw_language))

    language = language_manager.from_string(raw_language)
    try:
        search_results = api_manager.request_titles(author, language)
    except custom_exceptions.TitleNotFound:
        # logging_manager.error("Author not found: {}: {}".format(author, raw_language))
        raise

    answer = list(map(language_manager.transform_to_unicode, search_results))

    # logging_manager.info("API intent: {} author: {} raw_language: {} result: {}".format("search", author, raw_language, answer))
    return answer

def get_quotes(author, raw_language):
    # logging_manager.info("API intent: {} author: {} raw_language: {}".format("get_quotes", author, raw_language))
    language = language_manager.from_string(raw_language)

    suggested_author = search(author, raw_language)[0]
    quotes_page = api_manager.request_quotes_page(suggested_author, language)
    web_page_manager = html_manager.HTMLManager(quotes_page, language)

    web_page_manager.delete_sections(language.non_quote_sections)

    html_lists = web_page_manager.find_all_lists()
    quotes = []

    # Almost every quote structure is:
    # <ul>
    #   <li> Quote text </li>
    #   <ul>
    #       <li> author or some reference</li>
    #   </ul>
    # </ul>
    for html_list in html_lists:
        web_page_manager.remove_sublists(html_list)
        quotes.extend(html_manager.extract_text_from_list(html_list))

    answer = list(map(language_manager.transform_to_unicode, quotes))

    # logging_manager.info("API intent: {} author: {} resolved_author: {} raw_language: {} result: {}".format("get_quotes", author, suggested_author, raw_language, answer))

    return answer

def quote_of_the_day(raw_language):
    # logging_manager.info("API intent: {} raw_language: {}".format("quote_of_the_day", raw_language))
    language = language_manager.from_string(raw_language)
    quotes_page = api_manager.request_quote_of_the_day_page(language)
    web_page_manager = html_manager.HTMLManager(quotes_page, language)

    quote_of_the_day = language.quote_of_the_day_parser(web_page_manager.soup)
    quote_of_the_day = (language_manager.transform_to_unicode(quote_of_the_day[0]), language_manager.transform_to_unicode(quote_of_the_day[1]))

    # logging_manager.info("API intent: {} raw_language: {} result: {}".format("quote_of_the_day", raw_language, quote_of_the_day))

    return quote_of_the_day

def random_quote(author, raw_language):
    # logging_manager.info("API intent: {} author: {} raw_language: {}".format("random_quote", author, raw_language))

    random_quote = random.choice(get_quotes(author, raw_language))

    # logging_manager.info("API intent: {} author: {} raw_language: {} result: {}".format("random_quote", author, raw_language, random_quote))

    return random_quote

def supported_languages():
    # logging_manager.info("API intent: {}".format("supported_languages"))
    languages = ["english", "spanish"]
    return languages



import logging
import time
import inspect
from functools import wraps

logger = logging.getLogger("wikiquotes")
logger.addHandler(logging.NullHandler())

def error(message):
    logger.error(message)

def info(message):
    logger.info(message)

def debug(message):
    logger.debug(message)

def log_method_call(task):
    @wraps(task)
    def callable(*args, **kwargs):
        frame = inspect.stack()[1]
        module = inspect.getmodule(frame[0])
        function_called = "{}".format(task.__name__)
        logger.info("{} started".format(function_called))  # Use logger, not logging
        start_detection = time.time()
        result = task(*args, **kwargs)  # Call once and store result
        logger.info("{} finished: took {} seconds".format(function_called, time.time() - start_detection))  # Use logger, not logging
        return result  # Return the stored result
    return callable



from . import custom_exceptions

# Wikiquote api returns an html page inside the content ('extract')
def quotes_from_json(wikiquote_answer):
    pages = wikiquote_answer['query']['pages']

    text = ""

    for (key, value) in pages.items():
        try:
            text += pages[key]['extract']
        except KeyError:
            raise custom_exceptions.PageNotFoundException()

    return text

def correct_title_from_json(wikiquote_answer):
    search_info = wikiquote_answer['query']['searchinfo']

    try:
        suggestion = search_info['suggestion']
        return suggestion
    except KeyError:
        pass

    # If we reach here, no suggestions.
    search =  wikiquote_answer['query']['search']

    if search == []:
        raise custom_exceptions.TitleNotFound

    search_titles = []

    for search_result in search:
        search_titles.append(search_result['title'])

    return search_titles



class IncorrectAPIFormatException(Exception):
    pass

class PageNotFoundException(Exception):
    pass

class UnsupportedLanguageException(Exception):
    pass

class TitleNotFound(Exception):
    pass



from bs4 import BeautifulSoup, Tag

_titles = ['h2', 'h3', 'h4']

class HTMLManager:

    def __init__(self, html, language):
        self.language = language
        self.soup = BeautifulSoup(html, "lxml")

    # Although this doesn't need self, it mutates state so it's better to have it here and not as a free function.
    def remove_sublists(self, html_list):
            list_items = html_list.find_all('li')

            for li in list_items:
                 if li != -1 and li:
                    if li.ul:
                        li.ul.decompose()

    # Although this doesn't need self, it mutates state so it's better to have it here and not as a free function.
    def remove(self, element):
        for child in element.findChildren():
            child.decompose()

    def pretty_print(self):
        print(self.soup.prettify())

    def find_all_lists(self):
        return self.soup.find_all("ul")

    def delete_sections(self, non_quotes_section_keywords):

        def is_non_quote_section(html_title):
            for non_section_keyword in non_quotes_section_keywords:
                if non_section_keyword in html_title.text.lower():
                    return True
            return False

        all_sections = find_all_titles(self.soup)
        non_quotes_sections = list(filter(is_non_quote_section, all_sections))

        # Don't know exactly why, but can't decompose directly in loop. Have to do it later.
        elements_to_decompose = []

        for non_quotes_section in non_quotes_sections:
            for element in non_quotes_section.next_siblings:
                if is_title(element):
                    break
                else:
                    if isinstance(element, Tag) and element.name == "ul":
                        elements_to_decompose.append(element)

        # Deletes non quotes
        for element in elements_to_decompose:
            element.decompose()

        # This has to be done after deleting the quotes (sections aren't necesarily in order)
        for non_quotes_section in non_quotes_sections:
            non_quotes_section.decompose()

def is_title(html_element):
    return isinstance(html_element, Tag) and html_element.name in _titles

def find_all_titles(html):
    return html.find_all(_titles)

def extract_text_from_list(html_list):
    return map(lambda li_quote: li_quote.text.strip(), html_list.find_all("li"))



# -*- coding: utf-8 -*-
import unidecode

from ..languages import spanish
from ..languages import english
from . import custom_exceptions

def from_string(raw_language):
    lowered = strip_special_characters(raw_language.lower())
    if lowered == "en" or lowered == "english" or lowered == "ingles":
        return english
    elif lowered == "es" or lowered == "spanish" or lowered == "espanol":
        return spanish

    raise custom_exceptions.UnsupportedLanguageException()

def transform_to_unicode(string):
    return string

def strip_special_characters(accented_string):
    return unidecode.unidecode(accented_string)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

base_url = "https://es.wikiquote.org/w/api.php"
quote_of_the_day_url = "https://es.wikiquote.org/wiki/Portada"

def quote_of_the_day_parser(html):
    table = html("table")[1]("table")[0]
    quote = table("td")[3].text.strip()
    author = table("td")[5].div.a.text.strip()

    # Author and quote could have an accent
    return (quote, author)

non_quote_sections = ["sobre", "referencias"]
